/* $Id: jobline.cc,v 1.8 2003/04/27 15:57:21 richard Exp $ -*- C++ -*-
  __   _
  |_) /|  Copyright (C) 2003 Richard Atterer
  | \/|  <richard at atterer.net>
   '` 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2. See
  the file COPYING for details.

  One line in a JobList, in the lower part of the jigdo GUI window

*/

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <config.h>
#include <gtk-makeimage.hh>
#include <gtk-single-url.hh>
#include <jobline.hh>
#include <messagebox.hh>
#include <string.hh>
//______________________________________________________________________

namespace {

  /* Returns true if last characters of uri are the same as ext.
     Cannot use string::compare because of incompatibility between
     GCC2.95/3.0 */
  inline bool compareEnd(const string& uri, const char* ext) {
    static const unsigned extLen = strlen(ext);
    if (uri.size() < extLen) return false;
    unsigned pos = uri.size() - extLen;
    for (unsigned i = 0; i < extLen; ++i)
      if (ext[i] != uri[pos + i]) return false;
    return true;
  }

}
//______________________________________________________________________

void JobLine::create(const char* uri, const char* dest) {

  if (*uri == '\0') {
    MessageBox* m = new MessageBox(MessageBox::INFO, MessageBox::OK,
      _("Field for source URL/filename is empty"),
      _("Please enter an \"http\" or \"ftp\" URL to download, or the name "
        "of a <tt>.jigdo</tt> file to process."));
    m->show();
    return;
  }

  // Removing trailing '\' necessary on Windows
  string destination(dest);
  size_t lenLess1 = destination.length() - 1;
  if (lenLess1 > 0
      && destination[lenLess1] == DIRSEP)
    destination.resize(lenLess1);
  struct stat fileInfo;
  int statDest = stat(destination.c_str(), &fileInfo);
  bool destIsDir = S_ISDIR(fileInfo.st_mode);

  if (statDest != 0 && errno != ENOENT) {
    // destination is present, but error accessing it
    MessageBox* m = new MessageBox(MessageBox::ERROR, MessageBox::OK,
      _("Error accessing destination"),
      subst(_("The destination `%LE1' is present, but cannot be accessed: "
              "%LE2"), destination, strerror(errno)));
    m->show();
    return;
  }

  // Also create parent object
  size_t lastDirSep = destination.find_last_of(DIRSEP);
  string destParent(destination, 0, lastDirSep); // can be ==destination
  int statDestParent = stat(destParent.c_str(), &fileInfo);
  if (statDestParent != 0) {
    MessageBox* m = new MessageBox(MessageBox::ERROR, MessageBox::OK,
      _("Error accessing directory to save to"),
      subst(_("The destination `%LE1' cannot be accessed: %LE2"),
            destination, strerror(errno)));
    m->show();
    return;
  }

  /* We perform a regular download if the destination is either a filename
     (as opposed to a dir name), or the URI is not a jigdo file. */
  string uriStr(uri);
  if ((statDest != 0 && errno == ENOENT) // destination does not exist
      || (statDest == 0 && !destIsDir) // exists as file
      || !compareEnd(uriStr, ".jigdo")) { // URI end not ".jigdo"
    if (statDest == 0 && destIsDir) {
      // Append filename from source to directory name
      size_t lastDirSep = uriStr.find_last_of('/');
      destination += DIRSEP;
      /* If URL ends with /, use index.html as filename. DON'T try to use any
         name supplied by the server during a HTTP redirect - a malicious
         server could trick us into overwriting files! */
      if (lastDirSep == uriStr.length() - 1)
        destination.append("index.html");
      else
        destination.append(uriStr, lastDirSep + 1, string::npos);
    }
    GtkSingleURL* result = new GtkSingleURL(uriStr, destination);
    GUI::jobList.prepend(result);
    result->run();
    // NB result may already have deleted itself at this point
    return;
  }

# if !DEBUG
  MessageBox* m = new MessageBox(MessageBox::INFO, MessageBox::OK,
      "Work in progress",
      "Processing of .jigdo files is not yet implemented - only "
      "downloads of single files (e.g. <tt>.iso</tt> images) work.\n"
      "jigdo might crash any minute now - don't complain, I'm working on it! :-)");
  m->show();
# endif

  // Not a regular download => a .jigdo download
  GtkMakeImage* result = new GtkMakeImage(uriStr, destination);
  GUI::jobList.prepend(result);
  result->run();
  return;
}
//______________________________________________________________________

JobLine::~JobLine() {
  if (needTicks()) jobList()->unregisterTicks();
  jobList()->erase(row());
}
//______________________________________________________________________

void JobLine::waitTick() {
  if (--waitCountdown == 0) callRegularly(waitDestination);
# if DEBUG_JOBLIST
    cerr << "waitCountdown=" << waitCountdown << endl;
# endif
}
